#include "Robot.h"
#include "MedianFilter.h"
#include <NewPing.h>
#include <Pixy2.h>

#define RAD 3.14 / 180
#define GREY 750
#define MAX_DISTANCE 200 // Maximum distance (in cm) to ping.
#define PING_INTERVAL 35 // Milliseconds between sensor pings (29ms is about the min to avoid cross-sensor echo).
#define SONAR_NUM     3
#define ull unsigned long long int

/*
 * 0 bit - front
 * 1 bit - right
 * 2 bit - back
 * 3 bit - left
 */

MedianFilter filterDirection(3, 0);
Motor A (12, 42, 40);
Motor B (8, 46, 44);
Motor C (6, 34, 32);
Motor D (2, 38, 36);

IRLocator locator;
IRSeeker seeker;
Gyro gyro;
Pixy2 pixy;
Led ledL(11);
Led ledM(7);
Led ledR(3);

unsigned long pingTimer[SONAR_NUM]; // Holds the times when the next ping should happen for each sensor.
unsigned int cm[SONAR_NUM];         // Where the ping distances are stored.
unsigned int dis[SONAR_NUM];
uint8_t currentSensor = 0;

int x = 0;
int y = 0;
int alpha = 0;

bool stop_left = false;
bool stop_right = false;

int last_line = 0;
int speed1 = 255;
int hspeed = 315;

#include "forward.h"
#include "sensors.h"

void setup() {
  // put your setup code here, to run once:

  Serial.begin(9600);
  Serial3.begin(115200);
  
  locator.init();
  seeker.init();
  pixy.init();
  pixy.setLED(0, 0, 0);

  // Motors
  pinMode(10, OUTPUT);
  digitalWrite(10, HIGH);
  pinMode(4, OUTPUT);
  digitalWrite(4, HIGH);

  // Sensors
  pinMode(33, OUTPUT);
  pinMode(35, OUTPUT);

  // Buttons
  pinMode(13, INPUT_PULLUP);
  pinMode(9, INPUT_PULLUP);
  pinMode(5, INPUT_PULLUP);

  pinMode(11, OUTPUT);

  gyro.init(12000);

  pingTimer[0] = millis() + 75;           // First ping starts at 75ms, gives time for the Arduino to chill before starting.
  for (uint8_t i = 1; i < SONAR_NUM; i++) // Set the starting time for each sensor.
    pingTimer[i] = pingTimer[i - 1] + PING_INTERVAL;

}

void move(int speed, int angle){
  int error = gyro.read() * 3;
  
  A.start(-speed * cos((angle - 135) * RAD) + error);
  B.start(-speed * cos((angle + 135) * RAD) + error);
  C.start(-speed * cos((angle + 45) * RAD) + error);
  D.start(-speed * cos((angle - 45) * RAD) + error);

  int dx = speed * sin(angle * RAD) * 0.1;
  int dy = speed * cos(angle * RAD) * 0.1;
  x += dx;
  y += dy;
/*
  if (stop_right)
    ledR.on();
  else
    ledR.off();

  if (stop_left)
    ledL.on();
  else
    ledL.off();
    */
}


void back(bool side = true, int border = 30, int speed2 = speed1)
{
   stop();
   //ledM.on();
   int t = -90;
   int idx = 2;
   bool right = false;
   bool left = false;
  
   if (((line() >> 2) & 1) == 1 && ((line() >> 0) & 1) == 0)
   {
      ull t0 = millis();
      while (t0 + 100 >= millis())
      {
        move(speed2, 0);
      }
   }

   if (((last_line >> 1) & 1) == 1)
   {
     right = true;
     t = -90;
     idx = 2;
   }
   if (((last_line >> 3) & 1) == 1)
    {
       t = 90;
       idx = 0;
       left = true;
    }
    if (line() == 0)
    {
        left = false;
        right = false;
    }
   
   distance();
   if (side)
   {
       ull t1 = millis();
       if (right)
       {
         //ledR.on();
         while ((dis[2] < 50 || ((line() >> 1) & 1) == 1) && ((line() >> 3) & 1) != 1 && dis[0] >= 30 && t1 + 1000 >= millis())
         {
            move(speed1, -90);
            distance();
         }
       }
       else if (left)
       {
           //ledL.on();
           while ((dis[0] < 50 || ((line() >> 3) & 1) == 1) && ((line() >> 1) & 1) != 1 && dis[2] >= 30 && t1 + 1000 >= millis())
           {
              move(speed1, 90);
              distance();
           } 
           //ledL.off();
       }
       else
       {
          while (line() == 0 && (dis[0] < 50 || dis[2] < 50) && t1 + 400 >= millis())
          {
             move(speed2, t);
             distance();
          }
          if (line() != 0)
          {
            ull t3 = millis();
            while (t3 + 250 >= millis() && line() == 0)
            {
               move(speed2, -t);
            }
          }
          ledM.off();
       }
       stop_left = false;
       stop_right = false;
       x = 0;
       
   }

  ull t2 = millis();
  while ((dis[1] > border || dis[1] == 0) && ((line() >> 2) & 1) != 1 && t2 + 1300 >= millis())
  {
      move(speed2, 180);
      distance();
  }
  
  y = 0;
 

  if (((line() >> 2) & 1) != 0)
  {
    ledM.on();
    ull t0 = millis();
    while (t0 + 120 >= millis())
    {
      move(speed2, 0);
    }
    int bor = 30;
    int i = 0;
    if (((last_line >> 3) & 1) == 1)
    {
       t = 90;
       idx = 0;
       left = true;
    }
    if (((last_line >> 1) & 1) == 1)
    {
       t = -90;
       idx = 2;
       right = true;
    }
    distance();
    if (t == 90)
    {
      ull t0 = millis();
      //distance();
      while (((line() >> 1) & 1) == 0 && t0 + 400 >= millis())
      {
        move(speed2, 90);
        distance();
      }
      if (line() != 0)
      {
        ull cur = millis();
        while (dis[1] > 25 && cur + 300 >= millis() && (((line() >> 3) & 1) == 0))
        {
          move(speed2, -90);
          distance();
        }
      }
    }
    else
    {
      ull t0 = millis();
      while ((((line() >> 3) & 1) == 0 && t0 + 400 >= millis()))// (dis[1] > 25 || dis[1] == 0) &&
      {
        move(speed2, -90);
        distance();
      }
      if (line() != 0)
      {
        ull cur = millis();
        while (dis[1] > 25 && cur + 300 >= millis() && (((line() >> 1) & 1) == 0))
        {
          move(speed2, 90);
          distance();
        }
      }
    }
    stop();
    //stop_left = false;
    //stop_right = false;
    x = -t;
    ledM.off();
  }
  
}


int front_border = 50;
int side_border = 25; 
int cnt = 20;
int stay = 0;
int dir = 0;
ull tim = 0;


void goalie()
{
    while (line() == 0 && !digitalRead(5))
    {
      dir = locator.getDir1200();
      if (abs(dir) <= 90 && (abs(dir) <= 60 && seeker.getStr1200() > front_border || abs(dir) > 60 && locator.getStr1200() > side_border))
      {
        cnt = 20; // follow the ball!
      }
      else // probably stay
      {
        --cnt; 
        if (cnt < 0)
         cnt = 0;
      }  
      
      if (cnt > 0) // ball is coming
      {
        move(hspeed, dir); 
        stay = 0;
        stop_left = false;
        stop_right = false;
        if (tim + 3000 <= millis())
        {
          back();
          cnt = 0;
        }
        
        //ledM.off();
      }
      else // ball is far away from here
      {
          //ledM.on();
          tim = millis();
          
          distance();
          if (abs(dir) <= 10) // stay, pls
          {
            move(0, 0);
            back(false, 15, 250);
          }
          else if (dir > 0 && dir <= 110) // go right
          {
              stop_left = false;
              if (!stop_right)
              {
                  move(hspeed, 90);
              }
              else
              {
                 if (dis[1] > 25)
                 {
                    ull cur = millis();
                    stop_right = true;
                    while (dis[1] > 25 && cur + 400 >= millis() && line() == 0)
                    {
                      move(speed1, -90);
                      distance();
                    }
                    if (line() == 0)
                      stop_right = true;
                    stop(); 
                    //back(false, 15, 250);
                    if (dis[2] >= 65)
                      stop_right = false;
                 }
                 else
                 {
                   // if (dis[2] > 70)
                      //stop_right = false;
                    move(0, 0);
                 }
                   
              }
          }
          else if (dir < 0 && dir >= -110) // left side
          {
            stop_right = false;
            if (!stop_left)
            {
                move(hspeed, -90);
            }
            else
            {
               if (dis[1] > 25)
               {
                  ull cur = millis();
                  while (dis[1] > 25 && cur + 400 >= millis() && line() == 0)
                  {
                    move(speed1, 90);
                    distance();
                  }
                  if (line() == 0)
                    stop_left = true;
                  stop(); 
                 // back(false, 15, 250);
                  if (dis[0] >= 65)
                    stop_left = false;
                  
               }
               else
               {
                // if (dis[0] > 70)
                   //   stop_left = false;
                 move(0, 0);
               }
            }
          }
          else
          {
            move(0, 0);
            if (dis[1] > 50 || dis[1] == 0)
              back(false, 20, 250);
            else
              back(false, 15, 250);
          }
      }
      if (digitalRead(13) != 0)
      {
          while(digitalRead(13))
          {}
          Serial3.write(1);
          ledL.on();
          ledM.on();
          ledR.on();
          delay(100);
          ledL.off();
          ledM.off();
          ledR.off();
      }
   }
   stop();
   back();
}


void loop() {
/*
  while(true)
  {
    Serial.print(seeker.getStr1200());
    Serial.print(" ");
    Serial.println(locator.getStr1200());
  }
*/

   


  //while(digitalRead(9) == 0) {}


  while (!digitalRead(5))
  {
    goalie();
    ledR.on();
  }
  while(digitalRead(5)){}


  while (!digitalRead(5))
  {
    forward();
    ledR.off();
  }
  while(digitalRead(5)){}


 /*
  distance();
  Serial.print(dis[0]);
  Serial.print(" ");
  Serial.print(dis[1]);
  Serial.print(" ");
  Serial.print(dis[2]);
  Serial.println(" ");
  //Serial.println(angle);
  
  //int x = line();
  //ledL.on();

  
   //*/

   
   
/*
  Serial.print(locator.getDir1200());
  Serial.print(" ");
  Serial.print(locator.getDir600());
  Serial.print(" ");
  Serial.print(seeker.getDir1200());
  Serial.print(" ");
  Serial.print(seeker.getDir600());
  Serial.println(" ");
  */
  
  //Serial.println(locator.getDir1200());
  

  
  

  /*
   * following the ball
  while (true)
  {
    int cur = angle();
    if (cur > 130 || cur < -130)
      cur = 180;
    cur = cur * 1.3;
    move(100, cur);
    delay(1);
  }
  */
    

   /*
    * irseeker view
   int x1 = seeker.read1200();
   if (x1 > 127)
    x1 -= 254;
   Serial.print(x1);

   Serial.print(" ");
   Serial.print(locator.read1200() * 5);*/
   /*
   Serial.print(" ");
   Serial.print(seeker.getDir600());
   Serial.print(" ");
   Serial.print(locator.getDir1200());
   */
   
   /*
    * line sensors view
   int x = line();
   for (int i = 0; i < 4; ++i)
   {
       Serial.print(x % 2);
       x /= 2;
   }
   Serial.print("\n");
   delay(1);
   */
    
  

}
